<h3><a href="{{ pathto(master_doc) }}">{{ _('Table Of Contents') }}</a></h3>

<ul>
  <li><a href="{{ pathto('index') }}">Home</a></li>
  <li><a href="{{ pathto('widgets') }}">Widgets</a></li>
  <li><a href="{{ pathto('scripting_reference') }}">Scripting Reference</a></li>
</ul>

{%- if display_toc %}
  <hr>
  {{ toc }}
{%- endif %}