FTP_URL_ANNOTATIONS = 'ftp://ftp.ncbi.nlm.nih.gov/gene/DATA/gene2go.gz'
FTP_FILE_ANNOTATIONS = 'gene2go.gz'


FTP_URL_ONTOLOGY = 'http://www.geneontology.org/ontology/gene_ontology_edit.obo'

DOMAIN = 'go'
FILENAME_ANNOTATION = '{}.tab'


FILENAME_ONTOLOGY = 'gene_ontology.obo'
ONTOLOGY_TITLE = 'Gene Ontology (GO)'
ONTOLOGY_TAGS = ['gene', 'ontology', 'GO']
